/****************************************************************************\

**         XFDES.C - XTRA FAST DES ENCRYPTION ALOGRITHM FUNCTION            **

******************************************************************************

**             VAX/FDES Routines Modified By Doctor Dissector               **

******************************************************************************

** Copyright (c) 1991, By Doctor Dissector            Last Update: 07/24/91 **

\****************************************************************************/



/*

** XFDES - Version 6.00

** Copyright 1991 By Doctor Dissector

**

** This program is NOT free software BUT may be used without charge or

** payment in any form IF your copy is a "registered" distributed version.

** You may modify it as much as you please, however, you MAY NOT re-

** distribute it, in any shape or for: ie. modified OR unmodified,

** without the expressed written consent (ie. e-mail) of

** Doctor Dissector. (bbs.doctord@doomsday.spies.com)

**

** This program was initially distributed in the hope that it will be

** useful, but WITHOUT ANY WARRANTY; without even the implied warranty of

** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

*/



/*=[ Copyright ]============================================================*/



char xfdes_c_msg[] = "Source: xfdes.c, Copyright (c)1991 By Doctor Dissector";



/*=[ Include Files ]========================================================*/



#include "xfdes.h"



/*=[ Global Variables ]=====================================================*/



/*

** Key schedule. Alternating longs with low and high bits of key.

** Low and high 24 bits of keys are stored alternately.

*/

static sbpb24 KS[32];



/*

** It is slightly faster to indirect through this table than to specify

** the desired table directly.

*/

static sbpb24 *stt[] =

 {

  S0L, S0H,

  S1L, S1H,

  S2L, S2H,

  S3L, S3H,

  S4L, S4H,

  S5L, S5H,

  S6L, S6H,

  S7L, S7H,

 };



/* last password data */

static obpb1  oldblock[66];

static sbpb24 KSB[32];



#ifndef KC_XFDES

static unsb oldpw[21];

#endif



/*===========================================================================*/



/*

**    function: sbpb24 tfTOsixbit(ebpb24 tf)

** description: convert unsl in twenty-four bit contiguous format to six

**              bits per byte format

**      result: result of conversion

*/



sbpb24 tfTOsixbit(tf)

    ebpb24 tf;

{

    return((sbpb24) (((tf & 0x000003FL))        | ((tf & 0x0000FC0L) << 0x2) |

                     ((tf & 0x003F000L) << 0x4) | ((tf & 0x0FC0000L) << 0x6)));

}



/*===========================================================================*/



#ifndef KC_XFDES

/*

**    function: ebpb24 sixbitTOtf(sbpb24 sb)

** description: convert unsl in six bits per byte format to twenty-four

**              bit contiguous format

**      return: result of conversion

*/



ebpb24 sixbitTOtf(sb)

    sbpb24 sb;

{

    return((ebpb24) (((sb & 0x0000003FL))       | ((sb >> 0x2) & 0x0000FC0L) |

                     ((sb >> 0x4) & 0x003F000L) | ((sb >> 0x6) & 0x0FC0000L)));

}

#endif



/*===========================================================================*/



/*

**    function: void fsetkey(obpb1 *key)

** description: set up the key schedule from the key

**      return: none

*/



void fsetkey(key)

    obpb1 *key;

{

    reg    int   i, j, k;

    reg    obpb1 s, t;

    static obpb1 C[28], D[28];



    /*

    ** Generate C and D by permuting the key.  The low order bit of each

    ** 8-bit char is not used, so C and D are only 28 bits apiece

    */

    for(i=27; i>(-1); i--) {

        C[i]=key[PC1_C[i]];

        D[i]=key[PC1_D[i]];

    }

    /*

    ** Generate Ki by rotating C and D according to schedule and pick up

    ** a permutation using PC2

    */

    for(i=0; i<32; i+=2) {

        for(k=0; k<shifts[i]; k++) {     /* rotate */

            s=C[0];

            t=D[0];

            for(j=0; j<27; j++) {

                C[j]=C[j+1];

                D[j]=D[j+1];

            }

            C[27]=s;

            D[27]=t;

        }

        /* Get Ki, note C and D are concatenated */

        KS[i]=0;

        KS[i+1]=0;

        for(j=0; j<24; j++) {

            KS[i] |= ((sbpb24) C[PC2_C[j]] << j);

            KS[i+1] |= ((sbpb24) D[PC2_D[j]] << j);

        }

        KS[i]=tfTOsixbit(KS[i]);

        KS[i+1]=tfTOsixbit(KS[i+1]);

    }

}



/*===========================================================================*/



/*

**    function: void xform(sbpb24 *quarters, sbpb24 saltvalue)

** description: encrypt a block

**      return: none

*/



#ifndef NON_NETORDER

#define _AUTO_ (*dp--)

#define _OFFS_ 3

#else

#define _AUTO_ (*dp++)

#define _OFFS_ 0

#endif



void xform(quarters, saltvalue)

    sbpb24 *quarters, saltvalue;

{

    reg    int    i;

    reg    sbpb6  *dp;

    reg    sbpb24 **stp, *kp, k, Rl=0, Rh=0, Ll=0, Lh=0;

    static sbpb24 Dl, Dh, t;



    for(i=25; i>0; i--) {

        for(kp=KS; kp<(&KS[32]); ) {

            k=((Rl ^ Rh) & saltvalue);

            Dl=(k ^ (Rl ^ (*kp++)));

            Dh=(k ^ (Rh ^ (*kp++)));

            stp=((sbpb24 **) stt);

            dp=((sbpb6 *) &Dl) + _OFFS_;

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[_AUTO_];

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[_AUTO_];

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[_AUTO_];

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[(*dp)];

            dp=((sbpb6 *) &Dh) + _OFFS_;

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[_AUTO_];

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[_AUTO_];

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[_AUTO_];

            Ll ^= (*stp++)[(*dp)];

            Lh ^= (*stp++)[(*dp)];

            k=((Ll ^ Lh) & saltvalue);

            Dl=(k ^ (Ll ^ (*kp++)));

            Dh=(k ^ (Lh ^ (*kp++)));

            stp=((sbpb24 **) stt);

            dp=((sbpb6 *) &Dl) + _OFFS_;

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[_AUTO_];

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[_AUTO_];

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[_AUTO_];

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[(*dp)];

            dp=((sbpb6 *) &Dh) + _OFFS_;

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[_AUTO_];

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[_AUTO_];

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[_AUTO_];

            Rl ^= (*stp++)[(*dp)];

            Rh ^= (*stp++)[(*dp)];

        }

        t=Ll;

        Ll=Rl;

        Rl=t;

        t=Lh;

        Lh=Rh;

        Rh=t;

    }

#ifdef KC_XFDES

    (*quarters++)=(Ll | 0x21212121);

    (*quarters++)=(Lh | 0x21212121);

    (*quarters++)=(Rl | 0x21212121);

    (*quarters)=(Rh | 0x21212121);

#else

    (*quarters++)=sixbitTOtf(Ll);

    (*quarters++)=sixbitTOtf(Lh);

    (*quarters++)=sixbitTOtf(Rl);

    (*quarters)=sixbitTOtf(Rh);

#endif

}



/*===========================================================================*/



#ifndef KC_XFDES

/*

**    function: void toBA64F(reg sbpb24 *quarters, reg obpb1 *onebits64);

** description: creates table tmpE and executes final permutation

**      return: none

*/



void toBA64F(quarters, onebits64)

    sbpb24 *quarters;

    reg obpb1 *onebits64;

{

    reg    obpb1  *p, j;

    reg    sbpb24 quarter;

    static obpb1  tmpE[96];



    p=tmpE;

    quarter=(*quarters++);

    for(j=0; j<24; j++)

        (*p++)=((quarter >> j) & 0x1);

    quarter=(*quarters++);

    for(j=0; j<24; j++)

        (*p++)=((quarter >> j) & 0x1);

    quarter=(*quarters++);

    for(j=0; j<24; j++)

        (*p++)=((quarter >> j) & 0x1);

    quarter=(*quarters);

    for(j=0; j<24; j++)

        (*p++)=((quarter >> j) & 0x1);

    for(p=FP; (*p); )

        (*onebits64++)=tmpE[(*p++)];

}

#endif



/*===========================================================================*/



#ifdef KC_XFDES

/*

**    function: int kc_fcrypt(unsb *pw, sbpb24 smask, sbpb24 *in96, int nword)

** description: main encryption engine

**      return: 1 if match, 0 if no match

*/



int kc_fcrypt(pw, smask, in96, nword)

    unsb *pw;

    sbpb24 smask, *in96;

    int nword;

{

    reg    unsb   *tpw;

    reg    obpb1  i, j;

    static sbpb24 out96[4];

    static obpb1  block[66];



    if (nword) {

        for(i=0, tpw=pw; (*tpw) && (i<64); tpw++, i++) {

            for(j=0; j<7; j++, i++)

                block[i]=(((*tpw) >> (6-j)) & 0x1);

            block[i]=0;

        }

        for(; i<66;)

            block[i++]=0;

        fsetkey(block);

        for(i=0; i<32; i++) {

            KSB[i]=KS[i];

            oldblock[i]=block[i];

        }

        for(; i<66; i++)

            oldblock[i]=block[i];

    }

    else {

        for(i=0; i<32; i++) {

            KS[i]=KSB[i];

            block[i]=oldblock[i];

        }

        for(; i<66; i++)

            block[i]=oldblock[i];

    }

    xform(out96, smask);

    /* test pre-computed bits */

    if ((*in96++)==out96[0])

        if ((*in96++)==out96[1])

            if ((*in96++)==out96[2])

                if ((*in96)==out96[3])

                    return(1);

    return(0);

}



#else /*=====================================================================*/



/*

**    function: char *fcrypt(unsb *pw, char *salt)

** description: main encryption engine

**      return: encrypted <pw>

*/



char *fcrypt(pw, salt)

    unsb *pw;

    char *salt;

{

    reg    unsb   *tpw;

    reg    obpb1  c, i, j;

    static char   iobuf[16];

    static sbpb24 out96[4];

    static obpb1  block[66];



    if (strcmp((char *)pw, (char *)oldpw)) {

        for(i=0, tpw=pw; (*tpw) && (i<64); tpw++, i++) {

            for(j=0; j<7; j++, i++)

                block[i]=(((*tpw) >> (6-j)) & 0x1);

            block[i]=0;

        }

        for(; i<66; i++)

            block[i]=0;

        fsetkey(block);

        strcpy((char *)oldpw, (char *)pw);

        for(i=0; i<32; i++) {

            KSB[i]=KS[i];

            oldblock[i]=block[i];

        }

        for(; i<66; i++)

            oldblock[i]=block[i];

    }

    else {

        strcpy((char *)pw, (char *)oldpw);

        for(i=0; i<32; i++) {

            KS[i]=KSB[i];

            block[i]=oldblock[i];

        }

        for(; i<66; i++)

            block[i]=oldblock[i];

    }

    iobuf[0]=(*salt++);

    iobuf[1]=(*salt);

    xform(out96, tfTOsixbit((sbpb24) (SALT1[iobuf[0]] | SALT2[iobuf[1]])));

    toBA64F(out96, block);

    for(i=0; i<11; i++) {

        for(j=0, c=0; j<6; j++)

            c=((c << 0x1) | block[(6*i)+j]);

        iobuf[i+2]=CM[c];

    }

    iobuf[13]=0;

    if (!iobuf[1])

        iobuf[1]=iobuf[0];

    return(iobuf);

}

#endif



/*===========================================================================*/



#ifdef KC_XFDES

/*

**    function: void kc_decrypt(sbpb24 *in96, char *ct)

** description: partially decrypt the cyphertext <ct> into 96 bits to be

**              compared with the 96 bits OR'd with 0x8018017F after

**              the xform() function as called by kc_fcrypt()

**      return: nothing, but <in96> is modified to save the 96 bit values

*/



void kc_decrypt(in96, ct)

    sbpb24 *in96;

    char *ct;

{

    reg    int    i, j;

    reg    obpb1  *p, c;

           obpb1  block[66], tmpE[96];



    /* one char = 6 bits, store each bit seperately in array "block" */

    for(i=0; i<11; i++) {

        c=ICM[ct[i+2]];

        for(j=0; j<6; j++)

            block[(i*6)+j]=((c >> (5-j)) & 0x1);

    }

    /* set values to 1, you can then OR out the unused bits */

    for(i=0; i<96; i++)

        tmpE[i]=1;

    /*

    ** execute the final permutation backwards, placing values back

    ** into the tmpE array

    */

    for(i=0; i<64; i++)

        tmpE[FP[i]]=block[i];

    /* initialize the 4 parts of the 96 bit data to 0 */

    for(i=0; i<4; i++)

        in96[i]=0;

    /* now, for each of the 4 parts, set remaining 24 bits */

    for(i=96; i>72; i--)

        in96[3]=((in96[3] << 1) | (tmpE[i-1]));

    for(; i>48; i--)

        in96[2]=((in96[2] << 1) | (tmpE[i-1]));

    for(; i>24; i--)

        in96[1]=((in96[1] << 1) | (tmpE[i-1]));

    for(; i>0; i--)

        in96[0]=((in96[0] << 1) | (tmpE[i-1]));

    /*

    ** execute 24->sixbit conversion on each, so that it doesn't

    ** have to be done every time kc_fcrypt() is called

    */

    for(i=0; i<4; i++)

        in96[i]=tfTOsixbit(in96[i]);

}

#endif

