/****************************************************************************\
**      XFDES.H - FAST DES ENCRYPTION ALOGRITHM FUNCTIONS FOR TURBO-C       **
******************************************************************************
**             VAX/FDES Routines Modified By Doctor Dissector               **
******************************************************************************
** Copyright (c) 1991, By Doctor Dissector            Last Update: 03/04/91 **
\****************************************************************************/

/*
** XFDES - Version 1.05
** Copyright 1991 By Doctor Dissector
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 1, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

char xfdes_h_msg[]="xfdes v1.05 - 4/3/91 - (c)1991 - filename: xfdes.h";

/*=[ Definitions ]==========================================================*/

#define reg     register
#define uns     unsigned
#define unsb    uns     char
#define unsl    uns     long

/* Types for the different ways to represent DES bit patterns.
** Bits are always right justified within fields.
*/

#define obpb1   unsb    /* One bit per byte. */
#define sbpb6   unsb    /* Six bits per byte, 6 held. */
#define sbpb6R  unsb    /* Six bits per byte Reversed order, 6 held. */
#define sbpb24  unsl    /* Six bits per byte, 24 held. */
#define ebpb24  unsl    /* Eight bits per bit, 24 held. */
#define fbpb4   unsb    /* Four bits per byte, 4 held. */
#define fbpb4R  unsb    /* Four bits per byte Reversed order, 4 held. */

/*=[ Data Tables ]==========================================================*/

/* Final permutation, FP=IP^(-1)  - changed index from 1 to 0 */
unsb FP[] =
 {
  39, 7,47,15,55,23,63,31,
  38, 6,46,14,54,22,62,30,
  37, 5,45,13,53,21,61,29,
  36, 4,44,12,52,20,60,28,
  35, 3,43,11,51,19,59,27,
  34, 2,42,10,50,18,58,26,
  33, 1,41, 9,49,17,57,25,
  32, 0,40, 8,48,16,56,24,
 };

/* Permuted-choice 1 from the key bits to yield C and D.
** Note that bits 8,16... are left out: They are intended for a parity check.
** Not zero based!
*/
/* Modified 2/2/91 - DD */

static unsb PC1_C[] =
 {
  56,48,40,32,24,16, 8,
   0,57,49,41,33,25,17,
   9, 1,58,50,42,34,26,
  18,10, 2,59,51,43,35,
 };

static unsb PC1_D[] =
 {
  62,54,46,38,30,22,14,
   6,61,53,45,37,29,21,
  13, 5,60,52,44,36,28,
  20,12, 4,27,19,11, 3,
 };

/* Permuted-choice 2, pick out bits from CD array to generate key schedule. */
/* Modified 2/2/91 - DD */

static unsb PC2_C[] =
 {
  13,16,10,23, 0, 4,
   2,27,14, 5,20, 9,
  22,18,11, 3,25, 7,
  15, 6,26,19,12, 1,
 };

static unsb PC2_D[] =
 {
  12,23, 2, 8,18,26,
   1,11,22,16, 4,19,
  15,20,10,27, 5,24,
  17,13,21, 7, 0, 3,
};

/* The E bit-selection table. */
unsb E[] =
 {
  32, 1, 2, 3, 4, 5,
   4, 5, 6, 7, 8, 9,
   8, 9,10,11,12,13,
  12,13,14,15,16,17,
  16,17,18,19,20,21,
  20,21,22,23,24,25,
  24,25,26,27,28,29,
  28,29,30,31,32, 1,
 };

/* P is a permutation on the selected combination of the current L and key. */
static unsb P[] =
 {
  16, 7,20,21,
  29,12,28,17,
   1,15,23,26,
   5,18,31,10,
   2, 8,24,14,
  32,27, 3, 9,
  19,13,30, 6,
  22,11, 4,25,
 };

/* The 8 selection functions. */
static fbpb4R S[8][64] =
 {
  14, 4,13, 1, 2,15,11, 8, 3,10, 6,12, 5, 9, 0, 7,
   0,15, 7, 4,14, 2,13, 1,10, 6,12,11, 9, 5, 3, 8,
   4, 1,14, 8,13, 6, 2,11,15,12, 9, 7, 3,10, 5, 0,
  15,12, 8, 2, 4, 9, 1, 7, 5,11, 3,14,10, 0, 6,13,

  15, 1, 8,14, 6,11, 3, 4, 9, 7, 2,13,12, 0, 5,10,
   3,13, 4, 7,15, 2, 8,14,12, 0, 1,10, 6, 9,11, 5,
   0,14, 7,11,10, 4,13, 1, 5, 8,12, 6, 9, 3, 2,15,
  13, 8,10, 1, 3,15, 4, 2,11, 6, 7,12, 0, 5,14, 9,

  10, 0, 9,14, 6, 3,15, 5, 1,13,12, 7,11, 4, 2, 8,
  13, 7, 0, 9, 3, 4, 6,10, 2, 8, 5,14,12,11,15, 1,
  13, 6, 4, 9, 8,15, 3, 0,11, 1, 2,12, 5,10,14, 7,
   1,10,13, 0, 6, 9, 8, 7, 4,15,14, 3,11, 5, 2,12,

   7,13,14, 3, 0, 6, 9,10, 1, 2, 8, 5,11,12, 4,15,
  13, 8,11, 5, 6,15, 0, 3, 4, 7, 2,12, 1,10,14, 9,
  10, 6, 9, 0,12,11, 7,13,15, 1, 3,14, 5, 2, 8, 4,
   3,15, 0, 6,10, 1,13, 8, 9, 4, 5,11,12, 7, 2,14,

   2,12, 4, 1, 7,10,11, 6, 8, 5, 3,15,13, 0,14, 9,
  14,11, 2,12, 4, 7,13, 1, 5, 0,15,10, 3, 9, 8, 6,
   4, 2, 1,11,10,13, 7, 8,15, 9,12, 5, 6, 3, 0,14,
  11, 8,12, 7, 1,14, 2,13, 6,15, 0, 9,10, 4, 5, 3,

  12, 1,10,15, 9, 2, 6, 8, 0,13, 3, 4,14, 7, 5,11,
  10,15, 4, 2, 7,12, 9, 5, 6, 1,13,14, 0,11, 3, 8,
   9,14,15, 5, 2, 8,12, 3, 7, 0, 4,10, 1,13,11, 6,
   4, 3, 2,12, 9, 5,15,10,11,14, 1, 7, 6, 0, 8,13,

   4,11, 2,14,15, 0, 8,13, 3,12, 9, 7, 5,10, 6, 1,
  13, 0,11, 7, 4, 9, 1,10,14, 3, 5,12, 2,15, 8, 6,
   1, 4,11,13,12, 3, 7,14,10,15, 6, 8, 0, 5, 9, 2,
   6,11,13, 8, 1, 4,10, 7, 9, 5, 0,15,14, 2, 3,12,

  13, 2, 8, 4, 6,15,11, 1,10, 9, 3,14, 5, 0,12, 7,
   1,15,13, 8,10, 3, 7, 4,12, 5, 6,11, 0,14, 9, 2,
   7,11, 4, 1, 9,12,14, 2, 0, 6,10,13,15, 3, 5, 8,
   2, 1,14, 7, 4,10, 8,13,15,12, 9, 0, 3, 5, 6,11,
 };

/* Sequence of shifts used for the key schedule. */
static unsb shifts[] =
 {
  1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1,
 };
